import homepage from '@routes/index.html';
import auth from '@server/auth';
import ptApi from '@server/pt-api';
import favicon from '@static/favicon.png';
import robotsTxt from '@static/robots.txt';
import sitemapTxt from '@static/sitemap.txt';
import { env } from 'bun';

const development = env.NODE_ENV !== 'production';

const faviconInit = { headers: new Headers({ 'Content-Type': 'image/png' }) };

let ptEntriesETag = '';
let entriesETag = '';
let entriesCache = '';
let entriesCacheTime = 0;
const entriesCacheMaxAge = 3;
const entriesCacheTimeLimit = 1000 * entriesCacheMaxAge;

function getEtag(data: string) {
  const sha = Bun.SHA256.hash(data);
  return Buffer.from(sha.buffer).toString('base64url');
}

async function checkEntriesCache() {
  if (entriesCacheTime + entriesCacheTimeLimit >= Date.now())
    return;

  try {
    const [etag, data] = await ptApi.query(entriesETag);
    if (etag === ptEntriesETag || !Array.isArray(data)) return; // No new data
    if (etag !== null) ptEntriesETag = etag;

    entriesCache = JSON.stringify(data);
    entriesETag = getEtag(entriesCache);
    entriesCacheTime = Date.now();
  } catch (err) {
    console.error('Failed to fetch entries:', err);
  }
}

function etagResponse(etag: string, cacheControl?: string) {
  const headers = new Headers();
  headers.set('ETag', etag);
  if (cacheControl) headers.set('Cache-Control', cacheControl);
  return new Response('', { status: 304, headers });
}

Bun.serve({
  routes: {
    '/': homepage,
    '/robots.txt': new Response(robotsTxt, {
      headers: { 'Content-Type': 'text/plain' },
    }),
    '/sitemap.txt': new Response(sitemapTxt, {
      headers: { 'Content-Type': 'text/plain' },
    }),
    '/favicon.ico': new Response(favicon, faviconInit),
    '/health': new Response('OK'),
    '/api/entries': {
      async GET(req) {
        const jwt = await auth.verify(req.headers);
        if (jwt === null) {
          return auth.verifyFailResponse();
        }

        await checkEntriesCache();
        if (req.headers.get('if-none-match') === entriesETag) {
          return etagResponse(entriesETag, `max-age=${entriesCacheMaxAge}`);
        }

        return new Response(entriesCache, {
          headers: {
            'Content-Type': 'application/json',
            'Cache-Control': `max-age=${entriesCacheMaxAge}`,
            'ETag': entriesETag,
          }
        });
      },
    },
    '/auth/logout': {
      async POST() {
        return auth.logoutResponse();
      },
    },
    '/auth/verify': {
      async GET(req) {
        const jwt = await auth.verify(req.headers);
        if (jwt === null) return auth.verifyFailResponse(); // If no jwt token is found, return 401
        if (jwt === undefined) return new Response(); // If jwt is undefined it is still valid, return 200
        return auth.verifyResponse(jwt); // Renew jwt token
      },
    },
    '/auth/login': {
      async POST(req) {
        const json = await req.json();
        const data = json as { email?: string, password?: string };

        if (
          typeof data.email !== 'string' ||
          typeof data.password !== 'string' ||
          data.email.length < 3 ||
          data.email.indexOf('@') === -1 ||
          data.password.length === 0
        ) {
          return new Response('Incorrect email or password', { status: 400 });
        }

        const email = data.email.toLocaleLowerCase();
        const password = data.password;

        const jwt = await auth.login(email, password);
        if (!jwt) {
          return new Response('Incorrect email or password', { status: 400 });
        }

        return auth.loginResponse(jwt);
      },
    },
  },
  development,
  reusePort: true,
  // async fetch(req, server) {
  //   return new Response("Not found", { status: 404 });
  // },
});
