import { env } from 'bun';

export type SelectEntry = {
  id: number;
  name: string;
  href: string;
  finished: boolean;
  updated_at: string;
};

const baseUrl = env.API_SERVICE_URL;
const entriesUrl = `${baseUrl}/api/entries`;

const noBodyHeaders = new Headers({
  Authorization: `Bearer ${env.BEARER_TOKEN}`,
});

const bodyHeaders = new Headers({
  'Content-Type': 'application/json',
  Authorization: `Bearer ${env.BEARER_TOKEN}`,
});

async function query(etag?: string): Promise<[etag: string | null, data: Array<SelectEntry> | null]> {
  try {
    const headers = new Headers({
      Authorization: `Bearer ${env.BEARER_TOKEN}`,
    });

    if (etag) {
      headers.set('if-none-match', etag);
    }

    const response = await fetch(entriesUrl, {
      method: 'GET',
      signal: AbortSignal.timeout(5000),
      headers,
    });

    const respEtag = response.headers.get('etag');

    if (response.status === 304) {
      return [respEtag, null];
    }

    if (!response.ok) {
      throw new Error('Failed to fetch entries');
    }

    const data = await response.json();
    if (!Array.isArray(data)) {
      throw new Error('Invalid entries data');
    }

    return [respEtag, data];
  } catch (err) {
    console.error(err);
    throw new Error('Failed to fetch entries');
  }
}

async function remove(id: number) {
  try {
    const response = await fetch(`${entriesUrl}/${id}`, {
      method: 'DELETE',
      signal: AbortSignal.timeout(5000),
      headers: noBodyHeaders,
    });

    if (!response.ok) {
      throw new Error('Failed to remove entry');
    }
  } catch (err) {
    console.error(err);
    throw new Error('Failed to remove entry');
  }
}

async function update(id: number, body: { name: string; href: string }) {
  try {
    const response = await fetch(`${entriesUrl}/${id}`, {
      method: 'PUT',
      body: JSON.stringify(body),
      signal: AbortSignal.timeout(5000),
      headers: bodyHeaders,
    });

    if (!response.ok) {
      throw new Error('Failed to update entry');
    }
  } catch (err) {
    console.error(err);
    throw new Error('Failed to update entry');
  }
}

async function check(id: number, finished: boolean) {
  try {
    const response = await fetch(`${entriesUrl}/${id}/check/${finished}`, {
      method: 'PUT',
      signal: AbortSignal.timeout(5000),
      headers: noBodyHeaders,
    });
    if (!response.ok) {
      throw new Error('Failed to check entry');
    }
  } catch (err) {
    console.error(err);
    throw new Error('Failed to check entry');
  }
}

async function create(body: { name: string; href: string }) {
  try {
    const response = await fetch(`${entriesUrl}`, {
      method: 'POST',
      body: JSON.stringify(body),
      signal: AbortSignal.timeout(5000),
      headers: bodyHeaders,
    });
    if (!response.ok) {
      throw new Error('Failed to create entry');
    }
  } catch (err) {
    if (err) {
      console.error(err);
      throw new Error('Failed to create entry');
    }
  }
}

export default { query, remove, update, check, create };
